/*-
* renameFiles is a simple tool to rename many files at once
* Copyright (C) 2011 David Hobach
* tripleh [at] hackingthe [dot] net
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/

/** interface, that specifies the required methods of any QObjectReader for any input format (be it XML, binary or whatever...)
* since the input format is unknown in this interface, QByteArray is used
* some general methods are already implemented
* to write object data to some destination, use the corresponding writer to this reader
* NOTE: no use of any special QObject factories is enforced
* NOTE: signal + slot connections aren't currently supported; pointers within an object apart from the parent-child relationship and other objects within the object need some custom implementation --> this still has to be done by the programmer
* BUT: arbitrary QObject derived types and standard types within the class are supported; however they need to be used with the Q_PROPERTY macro and for reading custom types (self-defined classes and pointers), you need to derive QVariantHelper and use the Q_DECLARE_METATYPE macro for QVariant
* @see "Qt's Property System" in the documentation
* @since Qt 4.6
* @author David Hobach
* @version 0.3 / 11.04.2010
*/

/* maybe TODO:
see second note: it could be implemented by using some sort of id for each object and thereby identifying connections between objects
could become quite a big task though

test working with pointers
*/

#ifndef ABSTRACTQOBJECTREADER_H
#define ABSTRACTQOBJECTREADER_H

#include <QString>
#include <QByteArray>
#include <QObject>
#include <QList>

class AbstractQObjectReader
{

public:

/** read a QByteArray and construct a QObject from it
* @param QByteArray to read from
* @param parent QObject for the QObject that was read (optional)
* @return list of pointers to the constructed root QObjects (may have multiple subobjects) or an empty QList(), if the operation failed
*/
virtual QList<QObject*> fromBytes(const QByteArray& arr, QObject* parent = 0) = 0;

/** read from a file and construct a QObject from it; this basic implementation simply reads in binary mode and  then uses fromBytes()
* @param path to the file to read from
* @param parent QObject for the QObject that was read (optional); if set to NULL, the parent will be the internal factory, if possible; hence this reader should be destroyed correctly!!!
* @return list of pointers to the constructed root QObjects (may have multiple subobjects) or an empty QList(), if the operation failed
*/
virtual QList<QObject*> readFromFile(const QString& path, QObject* parent = 0);

/** virtual destructor for correct destruction of derived classes
*/
virtual ~AbstractQObjectReader() {;}

}; //end class

#endif
