/*-
* renameFiles is a simple tool to rename many files at once
* Copyright (C) 2011 David Hobach
* tripleh [at] hackingthe [dot] net
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/

/** interface, that specifies the required methods of any QObjectWriter for any output format (be it XML, binary or whatever...)
* since the output format is unknown in this interface, QByteArray is used
* some general methods are already implemented
* to read object data from some source, use the corresponding reader to this writer
* NOTE: signal + slot connections aren't currently supported; pointers within an object apart from the parent-child relationship and other objects within the object need some custom implementation --> this still has to be done by the programmer
* BUT: arbitrary QObject derived types and standard types within the class are supported; however they need to be used with the Q_PROPERTY macro and for reading custom types (self-defined classes and pointers), you need to derive QVariantHelper and use the Q_DECLARE_METATYPE macro for QVariant
* @see "Qt's Property System" in the documentation
* @since Qt 4.6
* @author David Hobach
* @version 0.3 / 11.04.2010
*/

/* maybe TODO:
see second note: it could be implemented by using some sort of id for each object and thereby identifying connections between objects
could become quite a big task though

test working with pointers
*/

#ifndef ABSTRACTQOBJECTWRITER_H
#define ABSTRACTQOBJECTWRITER_H

#include <QString>
#include <QByteArray>
#include <QObject>
#include <QStringList>

class AbstractQObjectWriter
{

public:

/** get a QByteArray representation of the specified QObject
* @param pointer to QObject to represent as QByteArray
* @param whether or not to inspect and recursively write child objects of the object to write (optional)
* @return QByteArray representation of the specified QObject; it has size 0, if the operation wasn't successful
*/
virtual QByteArray toBytes(const QObject* obj, bool writeChildren = true) = 0;

/** convenience function to write a QObject to a file; implemented to use toBytes() to write those bytes to the file
* @param path to the file to write the QObject to (overwrites existing files)
* @param pointer to the QObject to write to the file
* @param whether or not to inspect and recursively write child objects of the object to write (optional)
* @return true, if the operation was successful
*/
virtual bool writeToFile(const QString& path, const QObject* obj, bool writeChildren = true);

/** virtual destructor for correct destruction of derived classes in the case of polymorphism
*/
virtual ~AbstractQObjectWriter() {;}

protected: //some functions, that might be useful for implementing this class

/** get names of both dynamic and static properties from a QObject
* @param pointer to the QObject to retrieve all property names from
* @return QStringList containing the property names
*/
virtual QStringList propertyNames(const QObject* obj) const;

}; //end class

#endif
