/*-
* renameFiles is a simple tool to rename many files at once
* Copyright (C) 2011 David Hobach
* tripleh [at] hackingthe [dot] net
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/

/** macros to log events
 * there are many ways to do event logging
 * this file is meant to provide simple and common logging functionalities INDEPENDENTLY of the concrete mechanism used
 * basic idea: this file defines several macros that can be implemented for whatever logging class/API you want to use; some basic definitions are provided
 *
 * the following macros are defined for increasing severity: DBUG, INFO, WARN, ERR, ALERT, CRITICAL, EMERGENCY
 * note the funny spellings on DBUG and ERR to avoid conflicting with existing macros on Windows platforms
 * this file was inspired by the logog logger (http://johnwbyrd.github.com/logog/)
 *
 * the normal macros use a C++ style syntax (see examples below)
 * for each macro there exist _C versions with a C-like syntax
 * usually the C++ syntax is the better choice, because it works for ALL objects that implement the << operator
 *
 * usage examples:
 * INFO("Hello!");
 * WARN("Foo is > " << maxfoo << " Current value: " << foo << ".");
 * WARN_C("Foo is > %d! Current value: %d.", maxfoo, foo);
 *
 * the other macros defined here start with LM_
 * 
 * you can use the macros LM_DISABLE_DBUG, LM_DISABLE_WARN, ... to disable the corresponding log messages independently of the logging method
 * this will prevent any code from being generated for that log level
 *
 * options should be configured in the LogMacros_options.h file per project!
 * see LogMacros_options_example.h for an example
 *
 * NOTE: this file is not 100% portable, because Microsoft didn't implement vsnprintf according to the C99 and C++11 standard so far
 * (if you need this for MS compilers, use a trick such as in http://stackoverflow.com/questions/2915672/snprintf-and-visual-studio-2010)
 *
 * implementation note: if you need __VA_ARGS__ as string, use vsnprintf (no microsoft support, Microsofts _vsnprintf doesn't work correctly); see lm_simple_logging as example
 *
 * @author David Hobach
 * @version 0.3 / 16.06.12
 */

#ifndef LOGMACROS_H
#define LOGMACROS_H

#include "LogMacros_options.h" //this file is meant to contain YOUR concrete choice, which one of the following logging possibilities should be used (use "#define POSSIBILITY")

#include <cstdarg>
#include <cstdio>
#include <string>
#include <sstream>

//interprets the __VA_ARGS__ as format string and uses a SINGLE fprintf to write to the stream (thus it should be thread safe, because fprintf is atomic)
inline void lm_simple_logging(FILE* fp, const char* lvl, const char* file_line, const char* format, ...)
{
va_list ap;

va_start(ap, format);
//tip @buffer size from http://stackoverflow.com/questions/3919995/determining-sprintf-buffer-size-whats-the-standard
size_t size = vsnprintf(0,0,format,ap); //determine size
++size; //+1 for \0
char* fstr = new char[size];
vsnprintf(fstr,size,format,ap); //interpret format string
va_end(ap);

//print to FILE*
fprintf(fp,"%s : %s : %s\n",lvl,file_line,fstr);
fflush(fp);
delete fstr;
}

//macro definition --> string conversion as macro (LM_MACRO_TO_STRING); this works at compile time!
//see http://www.decompile.com/cpp/faq/file_and_line_error_string.htm
#define LM_STRINGIFY(x) #x
#define LM_MACRO_TO_STRING(x) LM_STRINGIFY(x)
#define LM_FILE_LINE __FILE__":"LM_MACRO_TO_STRING(__LINE__) //__LINE__ is an integer

//macro to convert a cpp string parameter (something such as "foo is " << foo << ".") to a c string (const char*) with the parameters being resolved (in our example: foo's value)
//note that the ostringstream object is unnamed to avoid naming conflicts
//the .flush() is necessary to obtain a non-const reference of the unnammed ostringstream (details: http://www.velocityreviews.com/forums/t729269-how-to-write-a-std-stringstream-one-liner.html)
//details:
//- std::stringstream() is a temporary, which cannot bind to a non-const reference. The << operator which takes a string literal is a non-member function whose first argument is a non-const reference.
//- The return value of << is always an std:stream&, not an std::stringstream&, so it doesn't have an str() function.
#define LM_CPP_TO_CSTRING(cppstr) ((std::ostringstream&)(std::ostringstream().flush() << cppstr)).str().c_str()

	//disable logging completely
	#ifdef LM_NO_LOGGING
		#define DBUG(cppstr)
		#define INFO(cppstr)
		#define WARN(cppstr)
		#define ERR(cppstr)
		#define ALERT(cppstr)
		#define CRITICAL(cppstr)
		#define EMERGENCY(cppstr)

		#define DBUG_C(...)
		#define INFO_C(...)
		#define WARN_C(...)
		#define ERR_C(...)
		#define ALERT_C(...)
		#define CRITICAL_C(...)
		#define EMERGENCY_C(...)

	//very simple logging to a FILE* defined by "#define LM_SIMPLE_LOGGING_FILEP <file pointer FILE*>"
	//note that stdout and stderr are file pointers
	//should be THREAD SAFE, but I'm not 100% sure --> to be tested
	#elif defined LM_SIMPLE_LOGGING
		#define LM_SIMPLE_LOGGING_EXEC(arg,cppstr) {fprintf(LM_SIMPLE_LOGGING_FILEP,"%s : %s : %s\n",arg,LM_FILE_LINE,LM_CPP_TO_CSTRING(cppstr)); fflush(LM_SIMPLE_LOGGING_FILEP);} //note the braces for locality
		#define LM_SIMPLE_LOGGING_EXEC_C(arg,...) lm_simple_logging(LM_SIMPLE_LOGGING_FILEP,arg,LM_FILE_LINE,__VA_ARGS__);

		#define DBUG(cppstr) LM_SIMPLE_LOGGING_EXEC("debug",cppstr)
		#define INFO(cppstr) LM_SIMPLE_LOGGING_EXEC("info",cppstr)
		#define WARN(cppstr) LM_SIMPLE_LOGGING_EXEC("warning",cppstr)
		#define ERR(cppstr) LM_SIMPLE_LOGGING_EXEC("error",cppstr)
		#define ALERT(cppstr) LM_SIMPLE_LOGGING_EXEC("alert",cppstr)
		#define CRITICAL(cppstr) LM_SIMPLE_LOGGING_EXEC("critical",cppstr)
		#define EMERGENCY(cppstr) LM_SIMPLE_LOGGING_EXEC("emergency",cppstr)

		#define DBUG_C(...) LM_SIMPLE_LOGGING_EXEC_C("debug",__VA_ARGS__)
		#define INFO_C(...) LM_SIMPLE_LOGGING_EXEC_C("info",__VA_ARGS__)
		#define WARN_C(...) LM_SIMPLE_LOGGING_EXEC_C("warning",__VA_ARGS__)
		#define ERR_C(...) LM_SIMPLE_LOGGING_EXEC_C("error",__VA_ARGS__)
		#define ALERT_C(...) LM_SIMPLE_LOGGING_EXEC_C("alert",__VA_ARGS__)
		#define CRITICAL_C(...) LM_SIMPLE_LOGGING_EXEC_C("critical",__VA_ARGS__)
		#define EMERGENCY_C(...) LM_SIMPLE_LOGGING_EXEC_C("emergency",__VA_ARGS__)

	//for usage with logog (http://johnwbyrd.github.com/logog/)
	//THREAD SAFE
	//(not tested)
	#elif defined LM_LOGOG
		#define LOGOG_USE_PREFIX 1 //use prefixes to avoid conflicts with our macros
		#include "logog.hpp"

		#define DBUG(cppstr) LOGOG_DEBUG("%s",LM_CPP_TO_CSTRING(cppstr))
		#define INFO(cppstr) LOGOG_INFO("%s",LM_CPP_TO_CSTRING(cppstr))
		#define WARN(cppstr) LOGOG_WARN("%s",LM_CPP_TO_CSTRING(cppstr))
		#define ERR(cppstr) LOGOG_ERROR("%s",LM_CPP_TO_CSTRING(cppstr))
		#define ALERT(cppstr) LOGOG_ALERT("%s",LM_CPP_TO_CSTRING(cppstr))
		#define CRITICAL(cppstr) LOGOG_CRITICAL("%s",LM_CPP_TO_CSTRING(cppstr))
		#define EMERGENCY(cppstr) LOGOG_EMERGENCY("%s",LM_CPP_TO_CSTRING(cppstr))

		#define DBUG_C(...) LOGOG_DEBUG(__VA_ARGS__)
		#define INFO_C(...) LOGOG_INFO(__VA_ARGS__)
		#define WARN_C(...) LOGOG_WARN(__VA_ARGS__)
		#define ERR_C(...) LOGOG_ERROR(__VA_ARGS__)
		#define ALERT_C(...) LOGOG_ALERT(__VA_ARGS__)
		#define CRITICAL_C(...) LOGOG_CRITICAL(__VA_ARGS__)
		#define EMERGENCY_C(...) LOGOG_EMERGENCY(__VA_ARGS__)
		
	#endif //LM_NO_LOGGING

/***************************************** macro implementation to disable logging of a certain level **************************************************/

	#ifdef LM_DISABLE_DBUG
		#undef DBUG
		#undef DBUG_C
		#define DBUG(cppstr)
		#define DBUG_C(...)
	#endif

	#ifdef LM_DISABLE_INFO
		#undef INFO
		#undef INFO_C
		#define INFO(cppstr)
		#define INFO_C(...)
	#endif

	#ifdef LM_DISABLE_WARN
		#undef WARN
		#undef WARN_C
		#define WARN(cppstr)
		#define WARN_C(...)
	#endif

	#ifdef LM_DISABLE_ERR
		#undef ERR
		#undef ERR_C
		#define ERR(cppstr)
		#define ERR_C(...)
	#endif

	#ifdef LM_DISABLE_ALERT
		#undef ALERT
		#undef ALERT_C
		#define ALERT(cppstr)
		#define ALERT_C(...)
	#endif

	#ifdef LM_DISABLE_CRITICAL
		#undef CRITICAL
		#undef CRITICAL_C
		#define CRITICAL(cppstr)
		#define CRITICAL_C(...)
	#endif

	#ifdef LM_DISABLE_EMERGENCY
		#undef EMERGENCY
		#undef EMERGENCY_C
		#define EMERGENCY(cppstr)
		#define EMERGENCY_C(...)
	#endif

#endif //LOGMACROS_H
