/*-
* renameFiles is a simple tool to rename many files at once
* Copyright (C) 2011 David Hobach
* tripleh [at] hackingthe [dot] net
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/

/** adds some mime type related features to the QFileInfo class
 * supports caching, but just per instance
 * @since Qt 4.6
 * @author David Hobach
 * @version 0.4 / 20.07.2011
 */

#ifndef QMIMEFILEINFO_H
#define QMIMEFILEINFO_H

#include <QFileInfo>
#include <QFile>
#include <QString>
#include <QDir>
#include <QByteArray>

class QMimeFileInfo : public QFileInfo {

public:

//ctors
QMimeFileInfo() : QFileInfo() {init(true);}
QMimeFileInfo(const QFile& file) : QFileInfo(file) {init(true);}
QMimeFileInfo(const QDir& dir, const QString& file) : QFileInfo(dir,file) {init(true);}
QMimeFileInfo(const QString& file) : QFileInfo(file) {init(true);}
QMimeFileInfo(const QFileInfo& fileinfo) : QFileInfo(fileinfo) {init(true);}

//copy ctor, doesn't copy the internal cache
QMimeFileInfo(const QMimeFileInfo& fileinfo) : QFileInfo(fileinfo) {init(true);}

//dtor
~QMimeFileInfo() {deleteCache();}

/** get the mime type of the file represented by this QFileInfo
 * NOTE: this function just has a rudimentary Linux implementation and will probably never be implemented on other systems
 * @param whether or not to include the encoding (optional)
 * @return mime type as string literal or an empty string in the case of an error
 */
QString getMimeType(bool includeEncoding = false) const;

/** checks whether the file represented by this QFileInfo is a binary
 * @return true, if it is a binary (must also be readable)
 */
bool isBinary();

/** checks whether the file represented by this QFileInfo is a text file
 * @return true, if it is a text file (must also be readable)
 */
bool isText();

/** specify whether methods such as isBinary() and isText() may use a cache for their work
 * @param true = cache is used (default)
 */
void setMimeCaching(bool useMimeCaching);

//override
void setFile(const QString& file) {deleteCache(); QFileInfo::setFile(file);}
void setFile(const QFile& file) {deleteCache(); QFileInfo::setFile(file);}
void setFile(const QDir& dir, const QString& file) {deleteCache(); QFileInfo::setFile(dir,file);}

private:
bool m_useCache;
QByteArray* m_cache; //== 0, if cache is not yet initialized

/** reads at least numBytes to the cache or doesn't do anything, if the cache contains those bytes
 * @param number of bytes to read (the cache may contain more due to previous reads)
 * @return the number of bytes available in the cache; -1 in the case of an error
 */
int readFirstBytesToCache(int numBytes = 100);

inline void deleteCache()
{
	if (m_cache)
	{
	delete m_cache;
	m_cache = 0;
	}
}

//helper methods
inline void init(bool useCache)
{
m_useCache = useCache;
m_cache = 0;
}

};

#endif
