/*-
* renameFiles is a simple tool to rename many files at once
* Copyright (C) 2011 David Hobach
* tripleh [at] hackingthe [dot] net
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/

/** general factory class, which supports creation of arbitrary QObject-derived classes, that have a default constructor
* advantages: single constructing function for all QObjects, QObjects are created with a parent --> no memory leaks, if the factory has a parent
* factory + singleton pattern
* non-virtual, because we don't need any subclasses?!
* NOTE: since we can't get a pointer to a constructor, we need this function template
* @author David Hobach
* @version 0.1 / 10.04.2010
* @since Qt 4.6
* @see http://bytes.com/topic/c/answers/760835-function-pointer-constructor
* @see http://stackoverflow.com/questions/957838/enforce-static-method-overloading-in-child-class-in-c
*/

#ifndef QOBJECTFACTORY_H
#define QOBJECTFACTORY_H

#include <QString>
#include <QObject>
#include <QHash>
#include <memory> //auto_ptr

namespace factories
{

template <typename T> QObject* create()
{
T* retT =  new T();
QObject* ret = static_cast<QObject*>(retT); //not dynamic, but at compile time!!! (for type checking of passed functions at compile time)
return ret;
}

typedef QObject* (*pfnConstructor)();

class QObjectFactory : public QObject //needs to be a QObject to be able to set the created objects' parent to this
{
Q_OBJECT

public:

/** get an instance of this factory
* @param set the parent of the instance (if set to NULL, it won't change the current parent; default parent is no parent = NULL) (optional); NOTE: it is recommended to set a parent, but the factory instance is also automatically deleted (and with it, all of its children)
* @return pointer to singleton instance of this factory
*/
static QObjectFactory* instance(QObject* parent = 0);

/** register a class to be used with this Factory
* NOTE: if the class was already registered before, the function pointer is overwritten
* @param name of that class to be used with build()
* @param pointer to a function without parameters, that can create objects of that class, which needs to be derived from QObject (optional)
*/
template <typename T> static void registerConstructor(const QString& className, QObject* (*pfnCreate)() = &create<T>) {s_map.insert(className,pfnCreate);} //templates need to be defined in the header file!

/** get an instance of a specified class
* @param the desired class to instantiate
* @param parent of the object to create (optional); if left out, the parent of the created QObject is set to this
* @return an instance of the desired class or NULL, if it couldn't be created (or casted to QObject*)
*/
QObject* build(const QString& className, QObject* parent=0);

/** Destructor, should never be called directly on instance()
*/
~QObjectFactory();

private:
static QHash<QString,pfnConstructor> s_map; //see typedef
static std::auto_ptr<QObjectFactory> s_instance; //use auto_ptr to make sure the pointer is deleted!!!

/** private Constructor
*/
QObjectFactory(QObject* parent);

/** private default Constructor
*/
QObjectFactory() {;}

//private copy ctor and assign operator should already be private by inheriting from QObject

};

} //end namespace

#endif
