/*-
* renameFiles is a simple tool to rename many files at once
* Copyright (C) 2011 David Hobach
* tripleh [at] hackingthe [dot] net
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/

/** container class which is a simple QObjectTree for QObject nodes
* this class simply employs QObjects built-in QObjectTree structure and provides some more features to make navigation within the QObjectTree easier
* each node may have an arbitrary number of subnodes
* the subnodes can only be seen, if we move to their parent node
* NOTE: it uses an implementation of the == operator for QObjects as hopefully specified in the template type implementation (for QObjects you can find one in QObjectSupport.h, to use pointer comparison simply pass a pointer type as T)
* trees can be saved as subtrees of this tree and so on... actually the tree structure of one QObject is added to this one, if it had one
* in general, functions with parameters that are passed by reference use == as defined by the template for equality, functions with pointers use pointer comparisons for equality
* @author David Hobach
* @version 0.4 / 08.04.2010
* @since Qt 4.6
*/

#ifndef QOBJECTTREE_H_
#define QOBJECTTREE_H_

#include <QObject>
#include <QList>

/*
TODO: 
testen, ob es mit einem MetaObjectParser geht (speichern + laden)
in libs packen
*/

class QObjectTree : public QObject
{
Q_OBJECT

public:

/** Constructor
* @param parent to use for this QObjectTree (without parent, it is necessary to explicitly destroy this QObjectTree or create it on the stack)
*/
QObjectTree(QObject* parent = 0);

/** destroys this QObjectTree AND all of its nodes (as usual with QObject)
*/
virtual ~QObjectTree(); //virtual so that the correct destructor gets called in the case of polymorphism (the one of the derived class)

//functions for QObjectTree navigation
/** get the root of this QObjectTree
* @return constant pointer to the root node, normally this
*/
virtual QObject* const getRoot(); //returns a constant pointer, not a pointer to a constant object!

/** find a node in the whole QObjectTree, that equals the given node (depends on the implementation of == of the template parameter)
* @param given node
* @return pointer to the node, that was found, NULL if it wasn't found (NOTE: only the first one found is returned)
*/
template <typename T> T* const find(const T& node)
{
return findR<T>(*this->getRoot(),node);
}

/** find all nodes in the whole QObjectTree, that equal the given node (depends on the implementation of == of the template parameter)
* @param given node
* @return list of pointers to the nodes, that were found to be equal to the given node or QList<T>(), if none was found
*/
template <typename T> const QList<T*> findAll(const T& node) //untested function
{
QList<T*> ret;
findAllR<T>(*this->getRoot(),node,ret);
return ret;
}

/** get the element within the QObjectTree we moved to last (the current parent element)
* @return pointer to the current parent element, that we can retrieve subnodes from
*/
virtual QObject* const getCurrent(); //return pointer instead of reference for easier usage of polymorphism

/** get the subnodes of the node, that is currently chosen as parent node by moveTo()
* @return const QObjectList containing those nodes
*/
virtual const QList<QObject*>& getCurrentSubNodes();

/** get all of the subnodes of type T of the node, that is currently chosen as parent node by moveTo()
* @return const QObjectList containing those nodes
*/
template <typename T> const QList<T*> getCurrentSubNodes() //note that a value is returned, not a reference!
{
QList<QObject*> children = this->getCurrentSubNodes();
QList<T*> ret;
QObject* cur;
T* casted;
	for (int i = 0; i < children.size(); ++i)
	{
	cur = children[i];
	casted = dynamic_cast<T*>(cur);
	if (casted == 0) continue;
	ret.append(casted);
	}
return ret;
}

/** sort all elements of type T in this QObjectTree based on the definition of their < operator
* the implementation employs a simple bubble sort optimized for almost sorted lists, in which case it might take almost linear time O(n); for unsorted lists O(n^2)
*/
template <typename T> void bSort()
{
bSortR<T>(*this->getRoot());
}

/** whether the current parent element has a subnode equal to the specified node
* NOTE: == from type T is used for equality
* @param node to compare with
* @return != 0, if the current parent has a subnode, which is equal to the one specified (requires impelemntation of ==); then the return value equals a pointer to that node
*/
template <typename T> T* const currentContains(const T& node)
{
QList<QObject*> children = this->getCurrentSubNodes();
QObject* cur;
T* casted;
	for (int i = 0; i < children.size(); ++i)
	{
	cur = children[i];
	casted = dynamic_cast<T*>(cur);
	if (casted == 0) continue;
	if (*casted == node) return casted;
	}
return 0;
}

/** add a subnode to the current parent node
* NOTE: the complete tree structure of that node is added to this tree!
* @param node to add (pointer!)
* @emit treeChanged()
*/
virtual void addSubNodeToCurrent(QObject* node); //NOTE: we cannot insert at specified positions due to the const limitations of getCurrentSubNodes()

/** remove a node from the tree (doesn't use any current parent)
* NOTE: in this case, pointer comparison is used for equality of QObjects
* @param node to remove
* @return pointer to the removed node (same as parameter), if the node did exist in the tree and could be removed (doesn't call the destructor), otherwise NULL
* @emit treeChanged() if tree is changed
*/
virtual QObject* const remove(QObject* node);

/** move to the specified node to use it as parent node (we can then see its children)
* NOTE: in this case, pointer comparison is used for equality of QObjects
* @param pointer to the node to move to
* @return whether the operation was successful or not
*/
virtual bool moveToPointer(QObject* node); //should have a different name than moveTo to prevent problems with polymorphism + templates

/** move to the specified node to use it as parent node (we can then see its children)
* NOTE: in this case, the template implementation of == is used for equality
* @param the node to move to
* @return whether the operation was successful or not
*/
template <typename T> bool moveTo(const T& node)
{
T* ret = find<T>(node);
if (!ret) return false;
this->current = ret;
return true;
}

/** set the current parent to the root
*/
virtual void reset();

/** whether this tree contains the specified pointer or not
* @param pointer to find in this tree
* @return true, if the tree does contain it
*/
virtual bool findPointer(const QObject* node);

signals:
/** is emitted whenever the tree is changed by using this interface
* @param the parent node where the tree was changed
*/
void treeChanged(QObject* parent);

protected:
QObject* current; //current parent
QObject* removed; //parent for the parentless QObjects that were removed

//template functions:

template <typename T> T* const findR(const QObject& start, const T& find)
{
QList<QObject*> children = start.children();
QObject *cur;
T *casted;
T *ret;
	for (int i = 0; i < children.size(); ++i)
	{
	cur = children[i];
	casted = dynamic_cast<T*>(cur);
	if ((casted != 0) && (*casted == find)) return cur;
	ret = findR<T>(*cur,find);
	if (ret != 0) return ret;
	}
return 0;
}

template <typename T> void findAllR(const QObject& start, const T& find, QList<T*>& list)
{
QList<QObject*> children = start.children();
if (children.isEmpty()) return;
QObject *cur;
T *casted;
	for (int i = 0; i < children.size(); ++i)
	{
	cur = children[i];
	casted = dynamic_cast<T*>(cur);
	if ((casted != 0) && (*casted == find)) list.append(casted);
	findAllR<T>(*cur,find,list);
	}
return;
}

template <typename T> void bSortR(const QObject& start)
{
const QList<QObject*> children = start.children(); //note the const'ness of the return value of children()
if (children.size() <= 0) return;
for (int i = 0; i < children.size(); ++i) bSortR<T>(*children[i]);

//get non-const copy of the items we need to sort:
QList<T*> cpy;
T* casted;
QObject* cur;
QObject* parent = children[0]->parent(); //representative parent for all of the children here
	for (int i = children.size()-1; i >= 0; --i)
	{
	cur = children[i];
	casted = dynamic_cast<T*>(cur);
		if (casted != 0)
		{
		casted->setParent(0); //basically remove it from the children list
		cpy.prepend(casted); //prepend to keep the order
		}
	}

//now we're using a slightly modified bubble sort to prevent more recursion; it should also work in nearly linear time, if the list was already sorted
bool noSwap; //for almost linear time in the case of a sorted list
T* swp;
	for (int i = 0; i < cpy.size()-1; ++i)
	{
	noSwap = true;
		for (int j = cpy.size()-1; j > i; --j)
		{
			if ((*cpy[j]) < (*cpy[j-1])) //requires implementation of the < operator for type T
			{
			noSwap = false;
			//swap (yes, you can also do that without the variable with xor, i know...)
			swp = cpy[j];
			cpy[j] = cpy[j-1];
			cpy[j-1] = swp;
			}
		}
	if (noSwap) break;
	}
	
//write the copy back
for (int i = 0; i < cpy.size(); ++i) cpy[i]->setParent(parent);
}

private:
bool findPointerR(const QObject* start, const QObject* find);

};

#endif
