/*-
* renameFiles is a simple tool to rename many files at once
* Copyright (C) 2011 David Hobach
* tripleh [at] hackingthe [dot] net
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/

#include "Tokenizer.h"
#include "Streams.h"

//#define TOKENIZER_DEBUG //switch on for debugging

void Tokenizer::init(QTextStream* inStream, const QRegExp& regex, int bufsize, const QString& ret, bool failOnRetMatch)
{
//actually initializing here is less performant than in the constructor, because this way it is initialized twice: once the constructor does the default init, then this function is called
//however the code is easier to read and modify this way

p_inStream = inStream;
m_regex = regex;
m_bufsize = bufsize;
m_err = false;
m_finished = false;
m_ret = ret;
m_failOnRetMatch = failOnRetMatch;
m_buffer = QString("");

//check for invalid input device
	if ((p_inStream->device() != 0) && (!p_inStream->device()->isReadable())) //strings are always readable (with device == 0)
	{
	cerr << "Tokenizer: input device not readable!" << endl;
	m_err = true;
	}

//check for invalid regex
	if (m_regex.captureCount() < 1)
	{
	cerr << "Tokenizer: specified invalid regex without any text captures!" << endl;
	m_err = true;
	}
}

Tokenizer::~Tokenizer()
{
if (p_inStream != 0) delete p_inStream;
}

void Tokenizer::fillBuffer()
{
	if (m_bufsize > 0)
	{
	char c;
		while ((m_buffer.size() < m_bufsize) && (!p_inStream->atEnd()) && (p_inStream->status() == QTextStream::Ok))
		{
		(*p_inStream) >> c;
		m_buffer.append(c);
		}
	}
	else m_buffer.append(p_inStream->readAll());
}

bool Tokenizer::addFirstBufferMatchToOutput()
{
int pos = -1;
pos = m_regex.indexIn(m_buffer,0);
//now we cheat a little bit: if pos != 0, we can still check for longer matches within this buffer by removing all characters until pos == 0
//note also, that we shouldn't simply remove all characters up to pos and refill the buffer as you might for example match (s1d)|(\d*[^a-z]) and the buffer is " s1" - so if we clear it now, we would have e.g. "1dk", which wouldn't match any of those 2
//moreover note, that pos is the position of the start of the MATCH and not the position of the first capture or something...
if ((pos != 0) && (!p_inStream->atEnd())) return false; //NOTE: special case p_inStream.atEnd() is needed: if it is at its end, we don't need out "cheat" anymore
if (pos < 0) return false;
bool ret = false;

QString capped;
	for (int i = 1; i < m_regex.captureCount()+1; ++i) //multiple text captures in one regex are possible
	{
	capped = m_regex.cap(i);
		if (capped != m_ret)
		{
		outQueue << capped; //with multiple captures, sometimes "" may be returned, if nothing was matched, but this isn't illegal (in the case of an (ads)|(k) for example: if ads is matched, the next thing capped is "", if k is not found) --> set m_ret to match your needs regarding your regex
		ret = true;
		}
		else
		{
			if (m_failOnRetMatch)
			{
			m_err = true;
			ret = false;
			}
		cerr << "Tokenizer: CRITICAL: a matched string equals ret --> no way to distinguish those two. Set a better ret in the Tokenizer constructor!" << endl;
		}
	}
pos += m_regex.matchedLength();
m_buffer.remove(0,pos); //remove match from buffer with pos being a counting variable
return ret;
}

bool Tokenizer::addAllBufferMatchesToOutput()
{
bool ret = false;
while (addFirstBufferMatchToOutput()) ret = true;
return ret;
}

QString Tokenizer::next()
{
//NOTE: calling addFirstBufferMatchToOutput only once and refilling the buffer directly afterwards makes sure, that matches always have the maximum possible length

	if (!m_finished)
	{
		while (outQueue.isEmpty())
		{
		fillBuffer();
		#ifdef TOKENIZER_DEBUG
		cout << "BUFFER: '" << m_buffer << "'" << endl; //for debugging
		#endif
		if (!addFirstBufferMatchToOutput()) m_buffer.remove(0,1); //shift 1 character out of the buffer, if nothing was found
			if ((m_err) || (p_inStream->atEnd()) || (p_inStream->status() != QTextStream::Ok))
			{
			addAllBufferMatchesToOutput();
			#ifdef TOKENIZER_DEBUG
			cout << "FINAL BUFFER: '" << m_buffer << "'" << endl; //for debugging
			#endif
			m_finished = true;
			break;
			}
		}
	}

if (!outQueue.isEmpty()) return outQueue.takeFirst(); //first is removed
else return m_ret;
}

bool Tokenizer::hasNext()
{
if (m_err) return false;

	if (outQueue.isEmpty())
	{
	if (m_finished) return false;
	QString nxt = next();
		if (nxt == m_ret) return false;
		else
		{
		outQueue.prepend(nxt);
		return true;
		}
	}
	else return true;
}

void Tokenizer::reset()
{
p_inStream->reset(); //see doc for details, why this doesn't do what is expected
p_inStream->resetStatus();
p_inStream->seek(0);
m_buffer.clear();
m_finished = false;
}

bool Tokenizer::nextBool(bool* ok)
{
if (ok != 0) *ok = true;
QString nxt = next().toLower();
	if ((nxt == "true") || (nxt == "1")) return true;
	if ((nxt == "false") || (nxt == "0")) return false;
if (ok != 0) *ok = false;
return false;
}

char Tokenizer::nextChar(bool* ok)
{
QString nxt = next();
	if (nxt.size() != 1)
	{
	if (ok != 0) *ok = false;
	return ' ';
	}
if (ok != 0) *ok = true;
return nxt.toLocal8Bit().at(0);
}

