/*-
* renameFiles is a simple tool to rename many files at once
* Copyright (C) 2011 David Hobach
* tripleh [at] hackingthe [dot] net
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/

/* class to create tokens defined by regular expressions from a text input stream
 * tokens are defined by regular expressions within enclosing braces "(" ")" - nested tokens are not supported as QRegExp doesn't support nested matching
 * basically this class just uses QRegExp::cap() on streamed input
 * example: using QRegExp rx("(\\d+)") as regex will result in tokens, which are positive integers
 * NOTE: Since we are reading from a stream, matching is performed against an internal buffer. If this buffer is full, characters at its beginning are discarded. This may result in some tokens not being matched, if (and only if) they are bigger than the size of the buffer. Therefore it is important to specify a buffer size, which is usable for your application or -1, if you know that your input stream is finite and don't care about memory consumption.
 * ALSO NOTE: multiple token definitions within one regex are supported
 * IMPORTANT: you shouldn't change the input of Tokenizer whilst using the class, if you don't know exactly what you're doing
 * this class is somewhat similar to the Java class "Scanner"
 * @see "QRegExp" in the documentation
 * @since Qt 4.6
 * @author David Hobach
 * @version 0.5 / 03.01.2011
 */

#ifndef TOKENIZER_H
#define TOKENIZER_H

#include <QTextStream>
#include <QRegExp>
#include <QIODevice>
#include <QByteArray>
#include <QString>
#include <QStringList>

class Tokenizer {

public:

//NOTE: the ctors are just analogous to those of QTextStream

/** ctor
 * @param device to read text from
 * @param regular expression to use for token matching
 * @buffer size; if -1 is specified, the whole input stream is read into a buffer before any matching is performed
 * @param string to be returned by next(), if an error occurred or the input stream is at its end
 * @param make Tokenizer fail, if ret equals a matched string; if set to false, the match is ignored and an error message printed to cerr
 */
Tokenizer(QIODevice* device, const QRegExp& regex, int bufsize, const QString& ret = "", bool failOnRetMatch = false) {init(new QTextStream(device),regex,bufsize,ret,failOnRetMatch);}

/** ctor
 * @param file handle to read text from
 * @param see above
 * @param see above
 * @param see above
 * @param see above
 */
Tokenizer(FILE* fileHandle, const QRegExp& regex, int bufsize, const QString& ret = "", bool failOnRetMatch = false) {init(new QTextStream(fileHandle,QIODevice::ReadOnly),regex,bufsize,ret,failOnRetMatch);}

/** ctor
 * @param QByteArray to read text from
 * @param see above
 * @param see above
 * @param see above
 * @param see above
 */
Tokenizer(QByteArray* array, const QRegExp& regex, int bufsize, const QString& ret = "", bool failOnRetMatch = false) {init(new QTextStream(array,QIODevice::ReadOnly),regex,bufsize,ret,failOnRetMatch);}

/** ctor
 * @param string to use as input stream
 * @param see above
 * @param see above
 * @param see above
 * @param see above
 */
Tokenizer(QString* string, const QRegExp& regex, int bufsize, const QString& ret = "", bool failOnRetMatch = false) {init(new QTextStream(string,QIODevice::ReadOnly),regex,bufsize,ret,failOnRetMatch);}

/** get the next token
 * @return next token or ret, if none was found or an error occurred; NOTE: "" may also be a valid match depending on the input regex, so you might have to choose a good ret value
 */
QString next();

/** check, whether a next token exists
 * @return true, if a next token exists
 */
bool hasNext();

/** whether the Tokenizer had some errors
 * @return true, if some error occurred
 */
bool hadError() {return m_err;}

/** resets the Tokenizer
 * */
void reset();

//setters and getters
void setRet(const QString& ret) {m_ret = ret;}
QString getRet() {return m_ret;}
QString getBuffer() {return m_buffer;}


/** convenience functions to get the next token as specified
 * @param is set to true, if everything went ok during the conversion and there was a next token; otherwise it is set to false (optional)
 * @return next token as the specified return value; may be invalid, if your regex doesn't ensure a String, that can be converted to the wanted return value
 */
int nextInt(bool* ok = 0) {return next().toInt(ok);}
unsigned int nextUInt(bool* ok = 0) {return next().toUInt(ok);}
long nextLong(bool* ok = 0) {return next().toLong(ok);}
float nextFloat(bool* ok = 0) {return next().toFloat(ok);}
double nextDouble(bool* ok = 0) {return next().toDouble(ok);}
short nextShort(bool* ok = 0) {return next().toShort(ok);}
bool nextBool(bool* ok = 0); //"1" and "true" are interpreted as true, "0" and "false" as false; other strings are interpreted as errors (case insensitive)
char nextChar(bool* ok = 0);

/** dtor
 */
~Tokenizer();

private:
QTextStream* p_inStream;
QString m_buffer;
QRegExp m_regex;
int m_bufsize;
bool m_err;
bool m_finished; //whether Tokenizer worked to the end of the stream or not
bool m_failOnRetMatch;
QString m_ret; //returned, if next() has nothing else to return
QStringList outQueue;

/** private init function used by constructors
 * @param pointer to a newly heap-allocated input stream to use
 * @param regex to use
 * @param maximum buffer size
 * @param return value to be used by next(), if an error occurred or the stream is at its end
 * @param whether or not to fail, if a match equals ret
 */
void init(QTextStream* inStream, const QRegExp& regex, int bufsize, const QString& ret, bool failOnRetMatch);

//fill buffer until its size is equal to m_bufsize
void fillBuffer();

//add first match within the buffer to the output queue and removes it from the buffer; returns true, if a match was added
bool addFirstBufferMatchToOutput();

//adds all matches found within the buffer to the buffer and removes them from there; returns true, if at least one match was added
bool addAllBufferMatchesToOutput();

}; //end class

#endif
