/* simple program to replace file names and/or file content by other file names/content as specified by a regular expression
 * not a real alternative to rename, but close (with some other features, but more dependencies ^^)
 * see usage()
 * @since Qt 4.6
 * @author David Hobach
 * @version 0.3 / 07.04.2011
 */

#include <QFileInfo>
#include <QFile>
#include <QTextStream>
#include <QDir>
#include <QList>
#include <QStringList>

#include "ArgumentList.h"
#include "Streams.h"
#include "TextFileReadWriteHelper.h"

void usage()
{
cout << "renamefiles <options> <regex to match> <string> <file or directory 1> ... <file or directory n>" << endl;
cout << "regex to match: extended regular expression as defined by the QRegExp class to be matched; capturing paranetheses are supported; all found matches are replaced by the given string" << endl;
cout << "string: string to use for replacement, may contain back-references such as \\1, \\2 and so on, which will be replaced by the content of the capturing parentheses" << endl;
cout << "file or directory: you may specify any number of files or directories (no recursion by default)" << endl;
cout << endl;
cout << "options:" << endl;
cout << "-r: recurse into the subdirectories of all given directories" << endl;
cout << "-F: rename all file names matching the given regex" << endl;
cout << "-C: replace all content of the given files matching the regex" << endl;
cout << "-f: overwrite existing files (only makes sense with the -F option)" << endl;
cout << "-v: verbose: print names of files, that were renamed and/or changed" << endl;
cout << "-n: no-act: show what files would have been renamed/changed, implies -v" << endl;
cout << "-w: use wildcard regular expressions instead of extended regex" << endl;
cout << "-i: case insensitive matching" << endl;
cout << "-A <regex>: only accept files with names matching the given regex as input for -C; affected by -i and -w; makes no sense without -C; the regex is evaluated BEFORE the renaming by -F is applied" << endl;
cout << "--charset <charset>: force usage of the given charset for reading and writing the text files (together with the -C option)" << endl;
cout << endl;
cout << "default behaviour: if neither -F nor -C are specified, -F is assumed" << endl;
}

/** retrieve arguments
 * @param ArgumentList to retrieve from; input parameter, that is changed: retrieved arguments are removed
 * @param regex to use; the -w option and -i option are already applied; output parameter
 * @param string; output parameter
 * @param list of files and directories to work on; output parameter
 * @param regex generated by the -A option; output parameter; set to ".*", if not used
 * @param output parameters
 * @return true, if all necessary parameters are given
 */
bool getArguments(ArgumentList& args, QRegExp* regex, QString* replace, QStringList* files, QRegExp* acceptFiles, bool* recurse, bool* renameFiles, bool* replaceContent, bool* forceOverwrite, bool* verbose, bool* noAct, QString* charset)
{
args.removeFirst(); //remove the current path
*recurse = args.getSwitch("-r");
*renameFiles = args.getSwitch("-F");
*replaceContent = args.getSwitch("-C");
*forceOverwrite = args.getSwitch("-f");
*verbose = args.getSwitch("-v");
*noAct = args.getSwitch("-n");
QString acceptFilesRegex = args.getSwitchArg("-A");
*charset = args.getSwitchArg("--charset","");
bool caseInsensitive = args.getSwitch("-i");
bool wildcard = args.getSwitch("-w");

if (*noAct) *verbose = true; //noAct implies -v

if (args.length() < 3) return false;

	if (acceptFilesRegex == "") acceptFiles->setPattern(".*");
	else acceptFiles->setPattern(acceptFilesRegex);

//create regex
regex->setPattern(args.takeFirst());
	if (wildcard)
	{
	regex->setPatternSyntax(QRegExp::Wildcard);
	acceptFiles->setPatternSyntax(QRegExp::Wildcard);
	}
	if (caseInsensitive)
	{
	regex->setCaseSensitivity(Qt::CaseInsensitive);
	acceptFiles->setCaseSensitivity(Qt::CaseInsensitive);
	}
	else
	{
	regex->setCaseSensitivity(Qt::CaseSensitive);
	acceptFiles->setCaseSensitivity(Qt::CaseSensitive);
	}

*replace = args.takeFirst();

for (int i = 0; i < args.length(); ++i) files->append(args[i]);

if ((!*renameFiles) && (!*replaceContent)) *renameFiles = true;

return true;
}

/** read the given text file (assumed to exist) and replace all content matching the given regex; errors are pinted to cerr
 */
void replaceFileContent(const QFileInfo& file, const QRegExp& regex, const QString& replace, bool verbose, bool noAct, const QString& charset)
{
bool ok;
TextFileReadWriteHelper tfhelper;
tfhelper.setCodec(charset);
QString content = tfhelper.readFile(file.absoluteFilePath(),&ok);
	if (!ok)
	{
	cerr << "Error whilst reading file '" << file.absoluteFilePath() << "'" << endl;
	return;
	}

bool changed = (regex.indexIn(content) != -1);

	if ((!noAct) && (changed))
	{
	content.replace(regex,replace);
	tfhelper.writeFile(file.absoluteFilePath(),content,&ok);
		if (!ok)
		{
		cerr << "Error whilst writing file '" << file.absoluteFilePath() << "'" << endl;
		return;
		}
	}

if ((verbose) && (changed)) cout << "Replace content: '" << file.absoluteFilePath() << "'" << endl;
}

/* real implemenentation of removeFile()
 */
bool removeFileImpl(const QFileInfo& file)
{
	if (file.isDir())
	{
	QDir dir(file.absoluteFilePath());
	bool ret;
	//DFS
	QList<QFileInfo> files = dir.entryInfoList(QDir::AllEntries | QDir::NoDotAndDotDot | QDir::System);
		for (int i = 0; i < files.size(); ++i)
		{
		ret = removeFileImpl(files[i]);
		if (!ret) return false;
		}
	//cout << "REMOVING dir " << file.absoluteFilePath() << endl;
	return dir.rmdir(file.absoluteFilePath()); //sometimes the Qt doc is just plain shit... rmdir seems to do what is necessary even though the doc says something like "name of the directory" != path
	} //usual file
	else
	{
	QFile toRemove(file.absoluteFilePath());
	//cout << "REMOVING file " << file.absoluteFilePath() << endl;
	return toRemove.remove();
	}
}

/** remove a file or directory (in the case of directories also remove all of their subelements)
 * @param file to remove
 * @param print verbose information
 * @param don't do anything, just print out information
 * @return true, if successful
 */
bool removeFile(const QFileInfo& file, bool verbose, bool noAct)
{
bool ret = true;
if (!noAct) ret = removeFileImpl(file);
if ((verbose) && (ret)) cout << "Remove: '" << file.absoluteFilePath() << "'" << endl;
return ret;
}

/** rename the given file (assumed to exist) according to the given regex; errors are directly printed to cerr
 * @return current file info (may be new or old, if nothing was done)
 */
QFileInfo renameFile(const QFileInfo& file, const QRegExp& regex, const QString& replace, bool forceOverwrite, bool verbose, bool noAct)
{
QString name, newName;
	if (file.isDir()) name = QDir(file.absoluteFilePath()).dirName();
	else name = file.fileName();
newName = name;
newName.replace(regex,replace);
if (name == newName) return file;

//different name
QFileInfo newFile(file.canonicalPath() + "/" + newName);
	if (newFile.exists())
	{
		if (!forceOverwrite)
		{
		cerr << "Couldn't rename file '" << file.absoluteFilePath() << "' to '" << newFile.absoluteFilePath() << "' - file or directory already exists." << endl;
		return file;
		}
		else
		{
			//delete the already existing file or directory
			if (!removeFile(newFile,verbose,noAct))
			{
			cerr << "-f: Couldn't completely remove file or directory '" << newFile.absoluteFilePath() << "'" << endl;
			return file;
			}
		}
	}

//now everything is settled and we can rename our file or directory
QDir parent = file.dir(); //get the parent directory
	if ((!noAct) && (!parent.rename(name,newName)))
	{
	cerr << "Couldn't rename file '" << file.absoluteFilePath() << "' to '" << newFile.absoluteFilePath() << "' - no access?" << endl;
	return file;
	}
	else
	{
		if (verbose)
		{
		cout << "Rename: '" << file.absoluteFilePath() << "' to '" << newFile.absoluteFilePath() << "'" << endl;
		}
	}
return newFile;
}

/** core function: rename the given file according to the given parameters; errors are directly printed to cerr
 * @param file to rename
 * @param regex to match against
 * @param string to insert
 * @param parameters
 */
void rename(const QFileInfo& file, const QRegExp& regex, const QString& replace, const QRegExp& acceptFiles, bool recurse, bool renameFiles, bool replaceContent, bool forceOverwrite, bool verbose, bool noAct, const QString& charset)
{
	if (!file.exists())
	{
	cerr << "Couldn't find file or directory: '" << file.absoluteFilePath() << "'" << endl;
	return;
	}

	if (file.isDir())
	{
	QDir dir(file.absoluteFilePath());
		if (recurse)
		{
		QList<QFileInfo> files = dir.entryInfoList(QDir::AllEntries | QDir::NoDotAndDotDot | QDir::System);
		//use DFS as we need correct directory names to work with
		for (int i = 0; i < files.size(); ++i) rename(files.at(i),regex,replace,acceptFiles,recurse,renameFiles,replaceContent,forceOverwrite,verbose,noAct,charset);
		} //else only rename the file name (below)
	}
	else //file
	{
	if ((replaceContent) && (acceptFiles.indexIn(file.fileName()) != -1)) replaceFileContent(file,regex,replace,verbose,noAct,charset);
	}

//in both cases
if (renameFiles) renameFile(file,regex,replace,forceOverwrite,verbose,noAct); //rename as second action as the file name might have changed otherwise for the first action
}

int main(int argc, char* argv[])
{
//QCoreApplication app(argc,argv); //NOTE: QApplication is for GUI apps, QCoreApplication for console apps! (if we use QApplication, fuzzy won't work without X server, e.g. in tty)
ArgumentList args(argc,argv);

QRegExp regex, acceptFiles;
QString replace, charset;
QStringList files;
bool recurse, renameFiles, replaceContent, forceOverwrite, verbose, noAct;

	if (!getArguments(args,&regex,&replace,&files,&acceptFiles,&recurse,&renameFiles,&replaceContent,&forceOverwrite,&verbose,&noAct,&charset))
	{
	usage();
	return 1;
	}
	
for (int i = 0; i < files.length(); ++i) rename(QFileInfo(files[i]),regex,replace,acceptFiles,recurse,renameFiles,replaceContent,forceOverwrite,verbose,noAct,charset);

return 0;
}
